# Copyright 2004-2005 Elemental Security, Inc. All Rights Reserved.
# Licensed to PSF under a Contributor Agreement.

"""Parser driver.

This provides a high-level interface to parse a file into a syntax tree.

"""

# Python imports
import os
import token
import logging
import tokenize

# Pgen imports
from pgen2 import parse, astnode, grammar

# Global objects
logger = logging.getLogger("pgen2.driver")

# Interface to load and cache the grammar

_grammar = None

def get_grammar():
    """Get the grammar, cached or freshly loaded."""
    global _grammar
    if _grammar is None:
        _grammar = load_grammar()
    return _grammar

here = os.path.dirname(__file__)
grammar_txt = os.path.join(here, "Grammar.txt")
grammar_pickle = os.path.join(here, "Grammar.pickle")

def load_grammar(gt=grammar_txt, gp=grammar_pickle, save=True, force=False):
    """Load the grammar (maybe from a pickle)."""
    if force or not newer(gp, gt):
        logger.info("Generating grammar tables from %s", gt)
        from pgen2 import pgen
        g = pgen.generate_grammar(gt)
        if save:
            logger.info("Writing grammar tables to %s", gp)
            g.dump(gp)
    else:
        g = grammar.Grammar()
        g.load(gp)
    return g

def newer(a, b):
    """Inquire whether file a was written since file b."""
    if not os.path.exists(a):
        return False
    if not os.path.exists(b):
        return True
    return os.path.getmtime(a) >= os.path.getmtime(b)


# Top-level parsing interface

def parse_stream_raw(stream, convert=None, debug=False):
    """Parse a stream and return the concrete syntax tree."""
    g = get_grammar()
    p = parse.Parser(g, convert)
    p.setup()
    t = v = x = None
    for t, v, x, y, z in tokenize.generate_tokens(stream.readline):
        if t in (tokenize.COMMENT, tokenize.NL):
            continue
        if t == token.OP:
            t = grammar.opmap[v]
        elif t == token.ERRORTOKEN:
            # This is a hack for a limitation of tokenize.py: it
            # doesn't recognize the '?' token, and somehow if one is
            # present in the source, it returns each whitespace
            # character before that as well as the '?' itself as a
            # separate error token.  Ignore the spaces and return the
            # token for the question mark
            if v.isspace():
                continue
            t = grammar.opmap.get(v, t)
        if debug:
            logger.debug("%s %r", token.tok_name[t], v)
        if p.addtoken(t, v, x):
            if debug:
                logger.debug("Stop.")
            break
    else:
        # We never broke out -- EOF is too soon (how can this happen???)
        raise parse.ParseError("incomplete input", t, v, x)
    return p.rootnode

def parse_stream(stream, debug=False):
    """Parse a stream and return the abstract syntax tree."""
    return parse_stream_raw(stream, astnode.convert, debug)

def parse_file(filename, debug=False):
    """Parse a file and return the abstract syntax tree."""
    stream = open(filename)
    try:
        return parse_stream(stream, debug)
    finally:
        stream.close()

def parse_string(text, debug=False):
    """Parse a string and return the abstract syntax tree."""
    from StringIO import StringIO
    stream = StringIO(text)
    return parse_stream(stream, debug)
